/*
 Vector ioDoc(
		    docCopyright("Steve Dekorte", 2002)
		    docLicense("BSD revised")
		    */

#include "IoVector_gl.h"
#include "List.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoOpenGL.h"
#include <math.h>

#include "GLIncludes.h"

#define DATA(self) ((IoVectorData *)IoObject_dataPointer(self))

void IoVector_glInit(IoObject *context) 
{
	IoState *state = ((IoObject *)context)->tag->state;
	IoObject *self = IoState_protoWithInitFunction_(state, IoVector_proto);
	
	{
		IoMethodTable methodTable[] = {
		{"glNormal", IoVector_glNormal},
		{"glTranslate", IoVector_glTranslate},
		{"glTranslatei", IoVector_glTranslatei},
		{"glRotate", IoVector_glRotate},
		{"glScale", IoVector_glScale},
		{"glUnproject", IoVector_glUnproject},
		{"glProject", IoVector_glProject},
			
		{"glVertex", IoVector_glVertex},
		{"glColor", IoVector_glColor},
		{"glClearColor", IoVector_glClearColor},
		{"drawQuadTo", IoVector_drawQuadTo},
		{"drawLineLoop", IoVector_drawLineLoop},
		{"drawLineLoopi", IoVector_drawLineLoopi},
		{"drawQuad", IoVector_drawQuad},
			
		{"red", IoVector_x},
		{"green", IoVector_y},
		{"blue", IoVector_z},
		{"alpha", IoVector_w},
			
		{"setRed", IoVector_setX},
		{"setGreen", IoVector_setY},
		{"setBlue", IoVector_setZ},
		{"setAlpha", IoVector_setW},
			
		{"drawAsLine", IoVector_drawAsLine},
		{"drawFilled", IoVector_drawFilled},
			
			/*
			 {"drawLineTo", IoVector_drawLineTo},
			 {"drawLineLoopTo", IoVector_drawLineLoopTo},
			 */
		{NULL, NULL},
		};
		IoObject_addMethodTable_(self, methodTable);
	}
}

IoObject *IoVector_glNormal(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glNormal3f(p->x, p->y, p->z); 
	return self; 
}

IoObject *IoVector_glTranslate(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glTranslatef(p->x, p->y, p->z); 
	return self; 
}

IoObject *IoVector_glTranslatei(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glTranslated(
			   (GLdouble)(GLint)(p->x), 
			   (GLdouble)(GLint)(p->y), 
			   (GLdouble)(GLint)(p->z)); 
	return self; 
}

IoObject *IoVector_glRotate(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	float angle = IoMessage_locals_doubleArgAt_(m, locals, 0);
	PointData *p = IoVector_rawPointData(self);
	glRotatef(angle, p->x, p->y, p->z); 
	return self; 
}

IoObject *IoVector_glScale(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glScalef(p->x, p->y, p->z); 
	return self; 
}

IoObject *IoVector_glVertex(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glVertex3f(p->x, p->y, p->z); 
	return self; 
}

IoObject *IoVector_glColor(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glColor4f(p->x, p->y, p->z, p->w); 
	return self; 
}

IoObject *IoVector_glClearColor(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	PointData *p = IoVector_rawPointData(self);
	glClearColor(p->x, p->y, p->z, p->w); 
	return self; 
}

IoObject *IoVector_glUnproject(IoVector *self, IoObject *locals, IoMessage *m)
{
	PointData *p = IoVector_rawPointData(self);
	GLint viewport[4];
	GLdouble mvmatrix[16];
	GLdouble projmatrix[16];
	GLdouble wx, wy, wz;
	
	glGetIntegerv(GL_VIEWPORT, viewport);
	glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
	
	gluUnProject((GLdouble)(p->x), 
			   (GLdouble)(p->y), 
			   (GLdouble)(p->z), 
			   mvmatrix, projmatrix, viewport, 
			   &wx, &wy, &wz);
	if (wz == 1) wz = 0; /* hack for orthographic projections? */
	
	(p->x) = wx;
	(p->y) = wy;
	(p->z) = wz;
	return self;
}

IoObject *IoVector_glProject(IoVector *self, IoObject *locals, IoMessage *m)
{
	PointData *p = IoVector_rawPointData(self);
	/* Project the receiver from object to window coordinates. */
	GLint viewport[4];
	GLdouble mvmatrix[16];
	GLdouble projmatrix[16];
	GLdouble objx, objy, objz;
	
	glGetIntegerv(GL_VIEWPORT, viewport);
	glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
	
	gluProject((GLdouble)(p->x), 
			 (GLdouble)(p->y), 
			 (GLdouble)(p->z),
			 mvmatrix, projmatrix, viewport,
			 &objx, &objy, &objz);
	
	(p->x) = objx;
	(p->y) = objy; 
	(p->z) = objz;
	return self;
}

IoObject *IoVector_drawLineTo(IoVector *self, IoObject *locals, IoMessage *m)
{
	IoVector *other = IoMessage_locals_pointArgAt_(m, locals, 0);
	PointData *p = IoVector_rawPointData(self);
	PointData *p2 = IoVector_rawPointData(other);
	glBegin(GL_LINES);
	glVertex3f((GLdouble)(p->x), (GLdouble)(p->y), (GLdouble)(p->z));
	glVertex3f((GLdouble)(p2->x), (GLdouble)(p2->y), (GLdouble)(p2->z));
	glEnd();
	return self;
}

IoObject *IoVector_drawLineLoopTo(IoVector *self, IoObject *locals, IoMessage *m)
{
	IoVector *other = IoMessage_locals_pointArgAt_(m, locals, 0);
	PointData *p = IoVector_rawPointData(self);
	PointData *p2 = IoVector_rawPointData(other);
	glBegin(GL_LINE_LOOP);
	glVertex3d((GLdouble)(p->x), (GLdouble)(p->y), 0);
	glVertex3d((GLdouble)(p2->x), (GLdouble)(p->y), 0);
	glVertex3d((GLdouble)(p2->x), (GLdouble)(p2->y), 0);
	glVertex3d((GLdouble)(p->x), (GLdouble)(p2->y), 0);
	glEnd();
	return self;
}

IoObject *IoVector_drawLineLoop(IoVector *self, IoObject *locals, IoMessage *m)
{
	PointData *p = IoVector_rawPointData(self);
	GLdouble x = (p->x);
	GLdouble y = (p->y);
	
	glBegin(GL_LINES);
	glVertex2d(x, y);
	glVertex2d(0, y);
	
	glVertex2d(0, y);
	glVertex2d(0, 0);
	
	glVertex2d(0, 0);
	glVertex2d(x, 0);
	
	glVertex2d(x, 0);
	glVertex2d(x, y);
	glEnd();
	return self;
}

IoObject *IoVector_drawLineLoopi(IoVector *self, IoObject *locals, IoMessage *m)
{
	PointData *p = IoVector_rawPointData(self);
	GLint x = (p->x);
	GLint y = (p->y);
	GLint s = 0;
	
	if (IoMessage_argCount(m)) 
	{ 
		s = (GLint)IoMessage_locals_doubleArgAt_(m, locals, 0); 
	}
	
	glBegin(GL_LINES);
	glVertex2i(x-s, y-s);
	glVertex2i(s, y-s);
	
	glVertex2i(s, y-s);
	glVertex2i(s, s);
	
	glVertex2i(s, s);
	glVertex2i(x-s, s);
	
	glVertex2i(x-s, s);
	glVertex2i(x-s, y-s);
	glEnd();
	return self;
}

IoObject *IoVector_drawQuadTo(IoVector *self, IoObject *locals, IoMessage *m)
{
	IoVector *other = IoMessage_locals_pointArgAt_(m, locals, 0);
	PointData *p = IoVector_rawPointData(self);
	PointData *p2 = IoVector_rawPointData(other);
	double x1, y1, x2, y2;
	double s = 0;
	
	if (IoMessage_argCount(m) > 1) 
	{ 
		s = IoMessage_locals_doubleArgAt_(m, locals, 1); 
	}
	
	x1 = (p->x) + s;
	y1 = (p->y) + s;
	x2 = (p2->x) - s;
	y2 = (p2->y) - s;
	
	glBegin(GL_QUADS);
	glVertex2d(x1, y1);
	glVertex2d(x2, y1);
	glVertex2d(x2, y2);
	glVertex2d(x1, y2);
	glEnd();
	return self;
}

IoObject *IoVector_drawQuad(IoVector *self, IoObject *locals, IoMessage *m)
{
	PointData *p = IoVector_rawPointData(self);
	double x, y, w, h;
	double s = 0;
	
	if (IoMessage_argCount(m) > 1) 
	{ 
		s = -IoMessage_locals_doubleArgAt_(m, locals, 1); 
	}
	
	x = s;
	y = s;
	w = (p->x) - (s*2);
	h = (p->y) - (s*2);
	
	glBegin(GL_QUADS);
	glVertex2d(w, h);
	glVertex2d(x, h);
	glVertex2d(x, y);
	glVertex2d(w, y);
	glEnd();
	return self;
}


/* --- vector stuff --- */

IoObject *IoVector_drawAsLine(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	Vector *v = IoVector_rawVector(self);
	int i, size = Vector_size(v);
	float *values = v->values;
	
	glBegin(GL_LINE_STRIP);
	
	for (i=0; i < size; i++)
	{
		float x = i;
		float y = values[i];
		
		glVertex2f(x, y);
	}
	
	glEnd();
	return self; 
}

IoObject *IoVector_drawFilled(IoVector *self, IoObject *locals, IoMessage *m)
{ 
	Vector *v = IoVector_rawVector(self);
	int i, size = Vector_size(v);
	float *values = v->values;
	
	glBegin(GL_TRIANGLE_STRIP);
	
	for (i=0; i < size; i++)
	{
		float x = i;
		float y = values[i];
		
		glVertex2f(x, 0);
		glVertex2f(x, y);
	}
	
	glEnd();
	return self; 
}

