/*
  callsystem    - system() on steorids

  features:
   - asynchonous running of a child process
   - setup of the environment
   - substitution of environment variables
   - connect all 3 standard streams to pipes, null devices, or files
   - pathname handling
   - #!shebang interpreter handling on any OS

  ALL IN A PORTABLE WAY!

  Note: the Win32 implementation hasnt yet the exact same semantic as the unix stuff,
        feel free to send me patches <chth@gmx.net>. The unix semantics are the mandatory semantics,
        don't change them!

  Pathnames:
    /   seperates a path,
    :   is used as device identifier when the path starts with / ("/C:/" becomes "C:\")
    .   is the current directory
    ..  is the parent directory

  Environment variables:
    only ${KEY} syntax is substituted in one pass, trying to substitute a non existing env var yields an error

  Note on error handling:
   Function returns follow the unix conventions which means:
     for int as return value:
       return 0 in case of success
       return -1 in case of an error
     for pointers as return value:
       return a valid pointer on success
       return NULL in case of an error

   To retrieve the actual error code one still needs to use the system specific method which is:
    on unix:  errno
    on Win32: GetLastError()   Note: errno is not set by Win32 API calls and is global (not thread local)

 */

#include "callsystem.h"


/*
  there is actually no way to tell if the forked process failed because of some
  callsystem trouble or because the called programm signaled an nonzero exit code,
  all we can do is to send an informative error message to stderr.
  In case of callsystem trouble the exitcode is always 127 *AND* a message formatted as
  with "CALLSYSTEM_ERROR: message : errnumber : error_as_string" is send to stderr.
*/
#ifdef unix
#define CALLSYSTEM_CHILD_ERROR(m)                                               \
 do{                                                                            \
  fprintf (stderr,                                                              \
           "\nCALLSYSTEM_ERROR: %s : %d : %s\n", m, errno, strerror(errno));    \
  exit (127);                                                                   \
 } while(0)
#elif defined(WIN32)
#define CALLSYSTEM_CHILD_ERROR(m) \
 child_die(m)
#endif



/*
  static declarations
 */
static char * alloc_executable_name (char ** env[], char ** argv[], const char * cmd);
static int veczsize(char ** vec[]);
static int setup_fd (callsystem_fd_t fds[2], callsystem_fd_t stdfd, int dir);
static size_t reserve_string (char ** s, size_t actual, size_t needed);

#ifdef WIN32
static void child_die (const char * msg);
static char* alloc_envblock (char * ptrs[]);
static char* alloc_commandline (char * ptrs[]);
#endif


/*
  int
  callsystem(const char * cmd,
             const char * argv[],
             const char * env[],
             callsystem_fd_t in[2],
             callsystem_fd_t out)[2],
             callsystem_fd_t err[2],
             const char * wd,
             int pri,
             callsystem_pid_t * child)

  parameters:
   cmd          command to execute
   argv         zero terminated array of options (only options,
                  without the usual argv[0] == cmd, callsystem handles that internally)
   envp         zero terminated array of environment vars
   (*in)[2]     fds to be connected to the childs stdin
   (*out)[2]    fds to be connected to the childs stdout
   (*err)[2]    fds to be connected to the childs stderr
   wd           working directory for the child process
   pri          priority advise 0=normal, 1=higher, -1=lower
   *child       where to store the childs pid, must be initialized to CALLSYSTEM_ILG_PID

  returns -1 when forking the child failed

  the fd's, argv, env and wd might be NULL
  envp will use defaults if NULL (see callsystem_exportdefaults() below)
  fd's will be bound to the standard streams when NULL
  working dir wont be changed if NULL
*/

//#include "/usr/include/sys/unistd.h"

int callsystem (const char * cmd,
            char * argv[],
            char * env[],
            callsystem_fd_t in[2],
            callsystem_fd_t out[2],
            callsystem_fd_t err[2],
            const char * wd,
            const int pri,
            callsystem_pid_t * const child)
{
#ifdef unix
  int ppri = getpriority(PRIO_PROCESS, 0);
  callsystem_pid_t cpid;

  if (*child != CALLSYSTEM_ILG_PID)
    {
      errno = EBUSY;
      return -1;
    }

  cpid = fork();
  if (cpid == -1)
    return -1;

  if (cpid == 0){
    /* child goes here */
    const char* bin;

    if ( setup_fd (in, STDIN_FILENO, 0) == -1)
      CALLSYSTEM_CHILD_ERROR("illegal stdin");

    if ( setup_fd (out, STDOUT_FILENO, 1) == -1)
      CALLSYSTEM_CHILD_ERROR("illegal stdout");

    if ( setup_fd (err, STDERR_FILENO, 1) == -1)
      CALLSYSTEM_CHILD_ERROR("illegal stderr");

    if (wd && chdir(wd))
      CALLSYSTEM_CHILD_ERROR("illegal working directory");

    if (!env && callsystem_exportdefaults (&env))
      CALLSYSTEM_CHILD_ERROR("callsystem_exportdefaults failed");

    bin = alloc_executable_name (&env, &argv, cmd);

    /* put command name into argv[0]*/
    callsystem_argv_pushfront (&argv, strrchr(cmd, '/')?(strrchr(cmd, '/') + 1):cmd);

    if (pri != 0){
      setpriority(PRIO_PROCESS, 0, ppri + (pri>0?-5:5));
      errno = 0;
    }

    execve(bin,argv,env);
    CALLSYSTEM_CHILD_ERROR("execve failed");
  }
  /* parent */
  /* close fd's */
  if (in)
      {
        close (in[0]);
        in[0] = CALLSYSTEM_ILG_FD;
      }
  if (out)
      {
        close (out[1]);
        out[1] = CALLSYSTEM_ILG_FD;
      }
  if (err)
      {
        close (err[1]);
        err[1] = CALLSYSTEM_ILG_FD;
      }

  *child = cpid;

  return 0;
#elif defined(WIN32)

  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPVOID argvblock;
  LPVOID envblock;

  ZeroMemory( &si, sizeof(si) );
  si.cb = sizeof(si);
  ZeroMemory( &pi, sizeof(pi) );

  /*TODO*/
  (void) cmd;
  (void) pri;
  (void) wd;
  (void) child;

  (void) setup_fd (in, STDIN_FILENO, 0);
  free(alloc_executable_name (&env, &argv, cmd));

  si.dwFlags = STARTF_USESTDHANDLES;


  /* TODO dup */

  if (!in)
      si.hStdInput = GetStdHandle (STD_INPUT_HANDLE);
  else
      si.hStdInput = in[0];

  if (!out)
      si.hStdOutput = GetStdHandle (STD_OUTPUT_HANDLE);
  else
      si.hStdOutput = out[1];

  if (!err)
      si.hStdError = GetStdHandle (STD_ERROR_HANDLE);
  else
      si.hStdError = err[1];

  /*appname*/



  /*commandline*/
  argvblock = alloc_commandline (argv);
  if (!argvblock)
    return -1;

  /*environment*/
  envblock = alloc_envblock (env);
  if (!envblock)
    return -1;




  /*
  if (!CreateProcess(
                     LPCTSTR lpApplicationName,
                     LPTSTR lpCommandLine,
                     NULL,
                     NULL,
                     FALSE,
                     0,
                     LPVOID lpEnvironment,
                     LPCTSTR lpCurrentDirectory,
                     &si,
                     &pi
                     ))
  */
    /*
    // Start the child process. 
    if( !CreateProcess( NULL,   // No module name (use command line). 
        TEXT("MyChildProcess"), // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi )             // Pointer to PROCESS_INFORMATION structure.
    ) 
    {
        printf( "CreateProcess failed (%d).\n", GetLastError() );
        return;
    }
    */
  /*TODO error and shebang handling (dont return here)*/
  return -1;

  /*
    BOOL SetPriorityClass(
    HANDLE hProcess,
  DWORD dwPriorityClass
  );

HIGH_PRIORITY_CLASS 	Process that performs time-critical tasks that must be executed immediately. The threads of the process preempt the threads of normal or idle priority class processes. An example is the Task List, which must respond quickly when called by the user, regardless of the load on the operating system. Use extreme care when using the high-priority class, because a high-priority class application can use nearly all available CPU time.
IDLE_PRIORITY_CLASS 	Process whose threads run only when the system is idle. The threads of the process are preempted by the threads of any process running in a higher priority class. An example is a screen saver. The idle-priority class is inherited by child processes.
NORMAL_PRIORITY_CLASS 	Process with no special scheduling needs.
  */


  /*
One way to obtain the current directory information for a drive X is to call GetFullPathName("X:",. .). That avoids an application having to scan the environment block. If the full path returned is X:\, there is no need to pass that value on as environment data, since the root directory is the default current directory for drive X of a new process.*/

  /*close other handles*/


#endif
}

/*
  int
  callsystem_running(callsystem_pid_t * pid);

  return >255 while the child is running,
  else its exit code
  return -1 in case of an error
*/
int
callsystem_running(callsystem_pid_t * pid)
{
#ifdef unix
  int status;

  if (*pid == CALLSYSTEM_ILG_PID)
    {
      errno = ECHILD;
      return -1;
    }

 redo:
  if (waitpid(*pid, &status, WNOHANG) == -1)
    {
      if ((errno = EINTR))
        {
          errno = 0;
          goto redo;
        }
      return -1;
    }

  if (!WIFEXITED(status))
    return 256;

  *pid = CALLSYSTEM_ILG_PID;
  return WEXITSTATUS(status);
#elif defined(WIN32)
/* http://msdn.microsoft.com/library/default.asp?url=/library/en-us/vccore98/html/_crt__cwait.asp
 */
  /*TODO*/
  (void) pid;
  return -1;
#endif
}

/*
  int
  callsystem_finished(callsystem_pid_t * pid);

  return -1 in case of error, else waits for the child
  finished and return its exit code
*/
int
callsystem_finished(callsystem_pid_t * pid)
{
#ifdef unix
  int status;

  if (*pid == CALLSYSTEM_ILG_PID)
    {
      errno = ECHILD;
      return -1;
    }

 redo:
  if (waitpid(*pid, &status, 0) == -1)
    {
      if ((errno = EINTR))
        {
          errno = 0;
          goto redo;
        }
      return -1;
    }

  *pid = CALLSYSTEM_ILG_PID;
  return WEXITSTATUS(status);
#elif defined(WIN32)
/* http://msdn.microsoft.com/library/default.asp?url=/library/en-us/vccore98/html/_crt__cwait.asp
 */
  /*TODO*/
  (void) pid;
  return -1;
#endif
}

/*
  int
  callsystem_terminate(callsystem_pid_t * );

  send the child a termination signal (might be handled or ignored)
  you still have to call callsystem_finished (or callsystem_running)
  for the child, but can't trust its return value.
*/
int
callsystem_terminate(callsystem_pid_t * pid)
{
#ifdef unix
  if (kill (*pid, SIGTERM) == -1)
    return -1;

  return 0;
#elif defined(WIN32)
  /*TODO*/
  (void) pid;
  return -1;
#endif
}

/*
  int
  callsystem_abort(callsystem_pid_t * );

  kill the child unconditionally (cant be handled or ignored)
  you still have to call callsystem_finished (or callsystem_running)
  for the child, but can't trust its return value.
*/
int
callsystem_abort(callsystem_pid_t * pid)
{
#ifdef unix
  if (kill (*pid, SIGKILL) == -1)
    return -1;

  return 0;
#elif defined(WIN32)
  /*TODO*/
  (void) pid;
  return -1;
#endif
}


/*
  fd's are always used as pairs (two element array)
  [0] is the used for reading
  [1] is used for writing

  in case of a pipe this fd's represent the sides of the pipe and callsystem()
  will close the proper sides automatically
  in case of other kinds of descriptors, either side will be CALLSYSTEM_ILG_FD and must not be accessed
*/


/*
  int
  callsystem_pipe(callsystem_fd_t pipes[2]);

  sets up a pipe
  returns -1 on error and sets errno
*/
int
callsystem_pipe(callsystem_fd_t pipes[2])
{
#ifdef unix
  return pipe (pipes);
#elif defined(WIN32)
  if (CreatePipe(&pipes[0],&pipes[1],NULL,0))
    return 0;
  else
    return -1;
#endif
}


/*
  int
  callsystem_null(callsystem_fd_t null[2]);

  portably opens a null device
  returns -1 on error and sets errno
*/
int
callsystem_null(callsystem_fd_t null[2])
{
#ifdef unix
  int fd = open("/dev/null",O_RDWR);
#elif defined(WIN32)
  HANDLE fd = CreateFile("NUL", GENERIC_READ | GENERIC_WRITE,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          0, OPEN_EXISTING, 0, NULL);
#endif
  if (fd == CALLSYSTEM_ILG_FD)
    return -1;

  null[0] = null[1] = fd;
  return 0;
}

/*
  int
  callsystem_open(const char * filename, enum callsystem_filemode mode,callsystem_fd_t (*fd)[2])

  open a file

  where mode is one of
   CALLSYSTEM_MODE_READ         - open *existing* file for reading
   CALLSYSTEM_MODE_WRITE        - not used here (dont use it)
   CALLSYSTEM_MODE_CREATE       - open *non existing* file for writing
   CALLSYSTEM_MODE_APPEND       - open *existing* file for writing at the end
   CALLSYSTEM_MODE_OVERWRITE    - open or create file for overwriting

  possibly or'ed with
   CALLSYSTEM_MODE_BINARY       - not evaluated here (but permitted)

  returns 0 on success, else -1 and sets errno
 */
int
callsystem_open(const char * filename,
                enum callsystem_filemode mode,
                callsystem_fd_t fd[2])
{
#ifdef unix
  switch (mode & ~CALLSYSTEM_MODE_BINARY)
    {
    case CALLSYSTEM_MODE_READ:
      fd[1] = CALLSYSTEM_ILG_FD;
      fd[0] = open(filename, O_RDONLY);
      if (fd[0] == -1)
        return -1;
      break;
    case CALLSYSTEM_MODE_CREATE:
      fd[0] = CALLSYSTEM_ILG_FD;
      fd[1] = open(filename, O_WRONLY|O_CREAT|O_EXCL);
      if (fd[1] == -1)
        return -1;
      break;
    case CALLSYSTEM_MODE_APPEND:
      fd[0] = CALLSYSTEM_ILG_FD;
      fd[1] = open(filename, O_WRONLY|O_APPEND);
      if (fd[1] == -1)
        return -1;
      break;
    case CALLSYSTEM_MODE_OVERWRITE:
      fd[0] = CALLSYSTEM_ILG_FD;
      fd[1] = open(filename, O_WRONLY|O_CREAT|O_TRUNC);
      if (fd[1] == -1)
        return -1;
      break;
    default:
      errno = EINVAL;
      return -1;
    }
#elif defined(WIN32)
  /*TODO*/
  (void) filename;
  (void) mode;
  (void) fd;
#endif
  return 0;
}

/*
  FILE *
  callsystem_fdopen(callsystem_fd_t * , enum callsystem_filemode mode);

  create a std C FILE stream from a underlying filedescriptor pair.

  fds   - filehandle pair
  mode  - the mode used in callsystem_open or
    CALLSYSTEM_MODE_READ
    CALLSYSTEM_MODE_WRITE
    for pipes

  possibly or'ed with
   CALLSYSTEM_MODE_BINARY       - open the file in binary mode

 */
FILE *
callsystem_fdopen(callsystem_fd_t fds[2], enum callsystem_filemode mode)
{
#ifdef unix
  switch (mode & ~CALLSYSTEM_MODE_BINARY)
    {
    case CALLSYSTEM_MODE_READ:
      return fdopen(fds[0], "r");
    case CALLSYSTEM_MODE_WRITE:
      return fdopen(fds[1], "w");
    case CALLSYSTEM_MODE_CREATE:
      return fdopen(fds[1], "w+");
    case CALLSYSTEM_MODE_APPEND:
      return fdopen(fds[1], "a");
    case CALLSYSTEM_MODE_OVERWRITE:
      return fdopen(fds[1], "w+");
    default:
      errno = EINVAL;
      return NULL;
    }
#elif defined(WIN32)
  /*TODO*/
  (void) fds;
  (void) mode;

  return NULL;
#elif 0
  switch (mode)
    {
      /*| CALLSYSTEM_MODE_BINARY */
    case CALLSYSTEM_MODE_READ:
      return fdopen(fds[0], "r");
    case CALLSYSTEM_MODE_WRITE:
      return fdopen(fds[1], "w");
    case CALLSYSTEM_MODE_CREATE:
      return fdopen(fds[1], "w+");
    case CALLSYSTEM_MODE_APPEND:
      return fdopen(fds[1], "a");
    case CALLSYSTEM_MODE_OVERWRITE:
      return fdopen(fds[1], "w+");
    default:
      errno = EINVAL;
      return NULL;
    }
#endif
}

/*
  int
  callsystem_close(callsystem_fds_t fds[2])

  close a file descriptor pair
*/
int
callsystem_close(callsystem_fd_t fds[2])
{
#ifdef unix
  int r=0;

  if (fds[0] == fds[1] && fds[0] != CALLSYSTEM_ILG_FD)
    {
      r += close (fds[0]);
    }
  else
    {
      if (fds[0] != CALLSYSTEM_ILG_FD)
        r += close (fds[0]);

      if (fds[1] != CALLSYSTEM_ILG_FD)
        r += close (fds[1]);
    }
  fds[0] = fds[1] = CALLSYSTEM_ILG_FD;

  if (r)
    return -1;
  else
    return 0;
#elif defined(WIN32)
  /*TODO*/
  (void) fds;
  return -1;
#endif
}



/*
  ARGV HANDLING

  These function take a pointer to a argv which must be initialized to NULL and allocate
  all memory which is needed. Strings are duped.

  An argv might be used for serveral callsystem invocations and must be freed after use.
 */

/*
  int
  callsystem_argv_pushback(char ** argv[], const char * const arg)

  push an arg onto the back of the argv
 */
int
callsystem_argv_pushback(char ** argv[], const char * const arg)
{
  void * tmp;
  size_t sz;

  if (!*argv)
    {
      *argv = malloc(sizeof(char*));
      if (!*argv)
        return -1;
      *argv[0] = NULL;
    }

  sz = veczsize(argv);

  tmp = realloc(*argv, (1 + sz) * sizeof(char*));
  if (!tmp)
    return -1;

  *argv = tmp;

  tmp = strdup (arg);
  if (!tmp)
    return -1;
  (*argv)[sz - 1] = tmp;
  (*argv)[sz] = NULL;

  return 0;
}

/*
  int
  callsystem_argv_pushfront(char ** argv[], const char * const arg)

  push an arg onto the front of the argv
 */
int
callsystem_argv_pushfront(char ** argv[], const char * const arg)
{
  void * tmp;
  size_t sz;

  if (!*argv)
    {
      *argv = malloc(sizeof(char*));
      if (!*argv)
        return -1;
      *argv[0] = NULL;
    }

  sz = veczsize(argv);

  tmp = realloc(*argv, (1 + sz) * sizeof(char*));
  if (!tmp)
    return -1;

  *argv = tmp;

  (void) memmove(*argv + 1, *argv, sz * sizeof (char*));

  tmp = strdup (arg);
  if (!tmp)
    return -1;
  (*argv)[0] = tmp;

  return 0;
}

/*
  free the argvector
 */
void
callsystem_argv_clear(char ** argv[])
{
  char ** i;

  if (!*argv)
    return;

  for (i=*argv;*i;++i)
    free(*i);

  free(*argv);
  *argv = NULL;
}


/*
  Setup of an environment for the child,
  this requires an initial char ** envp=NULL;
  such an environment must be freed after used (and might be reused in many callsystem calls)
 */

/*
  add an evnironment variable 'key' which holds 'value' to 'env'
 */
int
callsystem_setenv(char ** env[], const char * key, const char * value)
{
  void * tmp;
  size_t key_sz,sz;
  int envs;

  if (!*env)
    {
      *env = malloc(sizeof(char*));
      *env[0] = NULL;
    }
  if (!*env)
    return -1;

  sz = (key_sz = strlen (key)) + strlen (value) + 2;

  tmp = malloc (sz);
  if (!tmp)
    return -1;

  strcpy (tmp, key);
  ((char*)tmp)[key_sz] = '=';
  strcpy (((char*)tmp)+key_sz+1, value);

  (void) callsystem_unsetenv(env, key);

  envs = veczsize(env);
  (*env)[envs-1] = tmp;

  tmp = realloc(*env, (envs+1) * sizeof (char*));
  if (!tmp)
    {
      free((*env)[envs]);
      (*env)[envs] = NULL;
      return -1;
    }

  *env = tmp;
  (*env)[envs] = NULL;

  return 0;
}


/*
  return the value of environment variable 'key' from 'envp'
 */
const char *
callsystem_getenv(char ** env[], const  char * key)
{
  char ** i;
  size_t sz;

  sz = strlen (key);

  if (!env || !*env)
    return NULL;

  for (i = *env; *i; ++i)
    if (!strncmp(key,*i,sz) && (*i)[sz] == '=')
      return &(*i)[sz+1];

  return NULL;
}

/*
int
callsystem_unsetenv(char ** env[], const char * key)
  remove an environment variable 'key' from 'env'
  return 0 when successfully deleted
  return -1 in case 'key' was not part of the environment
 */
int
callsystem_unsetenv(char ** env[], const char * key)
{
  void * tmp;
  char ** i;
  size_t sz = strlen (key);

  if (!*env)
    return -1;

  for (i = *env; *i; ++i)
    if (!strncmp(key,*i,sz) && (*i)[sz] == '=')
      {
        free(*i);
        break;
      }

  if (!*i)
      return -1;

  while ((*i = *(i+1)))
    ++i;

  tmp = realloc(*env, (i - *env + 1) * sizeof (char*) );
  if (tmp)
    *env = tmp;

  return 0;
}

/*
  int
  callsystem_exportenv(char ** env[], const char * key)

  export a environment variable 'key' from the current environment to 'env'

  returns -1 on error
*/
int
callsystem_exportenv(char ** env[], const char * key)
{
  const char * val;

  val = getenv(key);
  if (!val)
    {
      errno = EINVAL;
      return -1;
    }

  return callsystem_setenv(env, key, val);
}

/*
  int
  callsystem_exportdefaults(char** env[])

  export the environment variables which are expected to be available on the current system

  return -1 on allocation error
 */
int
callsystem_exportdefaults(char** env[])
{
  static const char * callsystem_default_envvars[]={
#ifdef unix
    /* please review this list, I just added anything I found, naturally here shall only be the
       names which are expected for sane operation */
    "USER", "LOGNAME", "HOME", "LANG", "LC_ALL", "LC_COLLATE", "LC_CTYPE",
    "LC_MESSAGES", "LC_MONETARY", "LC_NUMERIC", "LC_TIME", "PATH", "MANPATH",
    "INFOPATH", "CDPATH", "DISPLAY", "PWD", "SHELL", "TERM", "PAGER", "EDITOR",
    "VISUAL", "BROWSER", "NLSPATH", "LOCPATH", "TMPDIR", "LD_LIBRARY_PATH",
    "LD_PRELOAD", "POSIXLY_CORRECT", "HOSTALIASES", "TZ", "TZDIR", "TERMCAP",
    "COLUMNS", "LINES",
#elif defined(WIN32)
  /*TODO*/
#endif
    0,
  };

  const char ** i;
  for ( i =  callsystem_default_envvars;*i;++i)
    {
      if (callsystem_exportenv(env, *i) == -1 && errno == ENOMEM)
        return -1;
    }
  return 0;
}

/*
  char *
  callsystem_env_subst(char ** envp[], const char * str);

  substitute environment variables from 'env' in 'str'
  return a freshly allocated string which must be freed after use

  environment variables must have the syntax ${KEY}, where KEY is the name of the variable
  A sole $ must be escaped by $$

  returns NULL on error and sets errno
      ENOMEM in case memory exhausted
      EINVAL in case of a syntax error, env var did not exist in environment

  Example: (environment FOO=foo BAR=bar)
    "${FOO}$$${BAR}"
  expands to:
    "foo$bar"
 */
char *
callsystem_env_subst(char ** env[], const char * str)
{
  char * ra;
  char * ret = NULL;
  size_t csr = 0;
  size_t sz = 0;

  for (; *str; ++str)
    {
      if (*str == '$')
        {
          ++str;
          if (*str == '{')
            {
              /* ${ */
              char * k;
              const char * v;
              size_t vs;
              const char * close;

              close = strchr (str + 1, '}');
              if (!close)
                goto einval;

              k = malloc (close - str);
              if (!k)
                goto emem;
              (void) strncpy (k, str + 1, close - str - 1);
              k[close - str - 1] = '\0';

              str = close;
              v = callsystem_getenv (env, k);
              free (k);

              if (!v)
                /* env var not found */
                goto einval;

              sz = reserve_string (&ret, sz, csr + (vs = strlen(v)) + 2);
              if (!sz)
                goto emem;

              (void) strcpy (&ret[csr], v);
              csr += vs;
            }
          else if (*str == '$')
            goto copy_char;
          else
            goto einval;
        }
      else
        {
        copy_char:
          sz = reserve_string (&ret, sz, csr + 1);
          if (!sz)
            goto emem;
          ret[csr++] = *str;
        }
    }

  ++csr;
  sz = reserve_string (&ret, sz, csr + 1);
  if (!sz)
    goto emem;
  ret[csr] = '\0';

  /* try to shrink the return space */
  ra = realloc (ret, sz);
  if (ra)
    ret = ra;

  return ret;

 einval:
  /* syntax error*/
  errno = EINVAL;
 emem:
  /* key allocation error */
  free (ret);
  return 0;
}


/*
  free all resources used by the environment 'env'
 */
void
callsystem_env_clear(char ** env[])
{
  char ** i;

  if (!*env)
    return;

  for (i=*env;*i;++i)
    free(*i);

  free(*env);
  *env = NULL;
}



/*
  char *
  callsystem_path_to_host(const char*)

  char *
  callsystem_path_to_unix(const char *)

  convert unix like filenames to OS specific filenames (and back)

  for win32:
   "/Letter:/" becomes "Letter:\"
   "/foo/" becomes "\foo\"

  returns a freshly allocated string or NULL on error
  must be freed after use
 */
char *
callsystem_path_to_host(const char * str)
{
#ifdef unix
  return strdup (str);
#elif defined(WIN32)
  /*TODO*/
  (void) str;
  return NULL;
#endif
}

char *
callsystem_path_to_unix(const char * str)
{
#ifdef unix
  return strdup (str);
#elif defined(WIN32)
  /*TODO*/
  (void) str;
  return NULL;
#endif
}



/*
  STATIC TOOL FUNCTIONS
 */

/*
  used internally, dup2 one fd and close the other of a fd pair
 */
static int
setup_fd (callsystem_fd_t fds[2], callsystem_fd_t stdfd, int dir)
{
#ifdef unix
  int flags;

  dir &= 1; /* only 0 an 1 are permitted */

  if (fds)
    {
      if (fds[dir] == CALLSYSTEM_ILG_FD)
        return -1;

    retry:
      if (dup2(fds[dir], stdfd) == -1)
        {
          if (errno == EINTR)
            goto retry;
          else
            return -1;
        }

      if ((flags = fcntl (stdfd, F_GETFD, 0)) < 0)
        return -1;
      if (fcntl (stdfd, F_SETFD, flags & ~FD_CLOEXEC) < 0)
        return -1;

      close (fds[dir]);
      close (fds[!dir]);
      fds[!dir] = fds[dir] = CALLSYSTEM_ILG_FD;
    }
  return 0;
#elif defined(WIN32)
  /*TODO*/
  (void) fds;
  (void) stdfd;
  (void) dir;
  return -1;
#endif
}

/*
  lookup the cmd in path, check access, check #! on windows and so on
 */
static char *alloc_executable_name(char ** env[], char ** argv[], const char * cmd)
{
#ifdef unix
  struct stat s;
  char * bin = NULL;

  (void) argv; /* not needed for unix, Win32 will need it implement #!/bin/... */

  if (strchr(cmd,'/') != NULL)
    {
      /* cmd includes path to binary */
      bin = strdup (cmd);
      if (!bin)
        CALLSYSTEM_CHILD_ERROR("strdup");

      if (stat (bin,&s) < 0 || !S_ISREG(s.st_mode))
        goto cmd_not_useable;
    }
  else
    {
      /* search cmd in $PATH */
      const char * path;
      char * path_dup;
      char * csr;
      char * pos;
      size_t csr_len;

      path = callsystem_getenv (env, "PATH");
      if (!path)
        CALLSYSTEM_CHILD_ERROR("PATH not set");

      path_dup = strdup (path);
      if (!path_dup)
        CALLSYSTEM_CHILD_ERROR("strdup");

      for (csr = strtok_r(path_dup, ":", &pos); csr; csr = strtok_r(NULL, ":", &pos))
        {
          bin = malloc ((csr_len = strlen (csr)) + strlen (cmd) + 2 );

          strcpy (bin,csr);
          bin[csr_len] = '/';
          strcpy (bin+csr_len+1, cmd);

          if (!stat (bin,&s) && S_ISREG(s.st_mode))
            break;

          free (bin);
          bin = NULL;
        }
      free (path_dup);
    }

  if (!bin)
    goto cmd_not_useable;

  if (access(bin, R_OK | X_OK) < 0)
    goto cmd_not_useable;

  return bin;
#elif defined(WIN32)
  /*TODO*/
  /*for OSes which dont support #!interpreters*/

  /*search cmd in PATH*/

  /*if check for #!shebang*/
  /*   search interpreter*/
  /*   check executable interpreter*/
  /*   check read cmd*/
  /*   assemble argv*/
  /*else*/
  /*   check executable cmd*/
  (void) env;
  (void) argv;
  (void) cmd;
  return NULL;

  goto cmd_not_useable;

#endif
 cmd_not_useable:
  CALLSYSTEM_CHILD_ERROR("cmd not executable");
}

/*
  count the members of a NULL terminated array of pointers
 */
static int
veczsize(char ** vec[])
{
  int i;
  char ** v = *vec;
  for (i=1;*v;++v,++i);
  return i;
}


/*
  realloc a string to at least needed size
  return the amount really reserved or 0 on error
 */
static size_t
reserve_string (char ** s, size_t actual, size_t needed)
{
  size_t n;
  char * r;

  for (n = actual>8?actual:8; n <= needed; n += (n>>1)); /*n = n * 1.5*/

  r = realloc (*s, n);
  if (!r)
    {
      /* that was to much, try conservatively */
      r = realloc (*s, n = needed);
      if (!r)
        return 0;
    }
  *s = r;
  return n;
}

/*
  WIN32 specific helpers
 */
#ifdef WIN32
static void
child_die (const char * msg)
{
  /*TODO*/
  (void) msg;
}

/*
  Win32 wants all env vars in a zero terminated block of C-Strings
  "foo=bar\0baz=barf\0\0" ...
 */
static char*
alloc_envblock (char * ptrs[])
{
  char ** csr;
  char * blk;
  char * blkcsr;
  size_t sz=0;

  if (!ptrs)
    return "";

  for (csr = ptrs; *csr; ++csr)
    {
      sz += (strlen (*csr) + 1);
    }

  blk = malloc (sz + 1);
  if (!blk)
    return NULL;

  blkcsr = blk;

  for (csr = ptrs; *csr; ++csr)
    {
      (void) strcpy(blkcsr,*csr);
      blkcsr += (strlen (*csr) + 1);
    }

  *blkcsr = '\0';
  return blk;
}

/*
  the commandline (argv) in Win32 is passed only as one string with
  proper double-quoting and escaping
 */
static char*
alloc_commandline (char * ptrs[])
{
  char ** csr;
  char * csr2;
  char * blk;
  char * blkcsr;
  size_t sz=0;

  if (!ptrs)
    return "";

  /* lets count the space needed*/
  for (csr = ptrs; *csr; ++csr)
    {
      sz += (strlen (*csr) + 1);

      /* spaces in string? we need 2 quotes */
      if (strchr (*csr, ' '))
        sz += 2;

      /* " have to be escaped, \" needs one extra char */
      for (csr2 = *csr; *csr2 && (csr2 = strchr (csr2, '"')); ++csr2)
          ++sz;

      /* i do NOT handle other whitespaces here, if that is a problem, send me patches */
    }

  blk = malloc (sz);
  if (!blk)
    return NULL;

  blkcsr = blk;

  /* now copy, quote, escape, terminate, space ... */
  for (csr = ptrs; *csr; ++csr)
    {
      /* leading " when needed */
      if (strchr (*csr, ' '))
        *blkcsr++ = '"';

      /* copy the string */
      for (csr2 = *csr; *csr2; ++csr2)
        {
          /* escape for \" */
          if (*csr2 == '"')
            *blkcsr++ = '\\';
          *blkcsr++ = *csr2;
        }

      /* trailing " when needed */
      if (strchr (*csr, ' '))
        *blkcsr++ = '"';

      /* one space to the next argument */
      *blkcsr++ = ' ';
    }

  /* trailing \0 overwrites the last space */
  *blkcsr = '\0';
  return blk;
}



#endif
  /*TODO*/
/*
  fprintf (stderr,                                                              \
           "\nCALLSYSTEM_ERROR: %s : %d : %s\n", m, errno, strerror(errno));    \
  exit (127);                                                                   \

*/
/*
void ErrorExit(LPTSTR lpszFunction) 
{ 
    TCHAR szBuf[80]; 
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    wsprintf(szBuf, 
        "%s failed with error %d: %s", 
        lpszFunction, dw, lpMsgBuf); 
 
    MessageBox(NULL, szBuf, "Error", MB_OK); 

    LocalFree(lpMsgBuf);
    ExitProcess(dw); 
}

*/
