/* Copyright (c) 2004, Kentaro A. Kurahone
 * All Rights Reserved.
 */

#include "IoState.h"
#include "IoObject.h"
#include "IoNil.h"
#include "IoArray.h"
#include "IoNumber.h"
#define DATA(self) ((IoArrayData *) self->data)

/* Static methods... */
static IoTag* IoArray_tag(void *state);
static IoObject* IoArray_handleAt(IoArray *self, IoObject *locals, IoMessage *m, int flags);

static IoTag* IoArray_tag(void *state)
{
  IoTag *tag = IoTag_newWithName_("Array");
  tag->state = state;
  tag->cloneFunc = (TagCloneFunc *) IoArray_rawClone;
  tag->markFunc = (TagMarkFunc *) IoArray_mark;
  tag->freeFunc = (TagFreeFunc *) IoArray_free;
  return tag;
}

static IoObject* IoArray_handleAt(IoArray *self, IoObject *locals, IoMessage *m, int at)
{
  IoObject *val;
  IoObject *newVal;
  long i;
  long d;
  long j;

  if (DATA(self)->d == 0 || 
      DATA(self)->size == NULL ||
      DATA(self)->idxCache == NULL ||
      DATA(self)->backingStore == NULL)
  {
    if (at == 1)
      IoState_error_description_(IOSTATE, m, "Array.at",
        "Array not initialised");
    else
      IoState_error_description_(IOSTATE, m, "Array.atPut",
        "Array not initialised");
  }

  if (at == 1) 
  {
    if (IoMessage_argCount(m) < DATA(self)->d)
      IoState_error_description_(IOSTATE, m, "Array.at",
        "Not enough args");
  }
  else
  {
    if (IoMessage_argCount(m) < DATA(self)->d + 1)
      IoState_error_description_(IOSTATE, m, "Array.atPut",
        "Not enough args");
  }

  i = 0;
  for (d = 0; d < DATA(self)->d; d++)
  {
    j = IoMessage_locals_longArgAt_(m, locals, d);
    if (j > DATA(self)->size[d] - 1)
    {
      if (at == 1)
        IoState_error_description_(IOSTATE, m, "Array.at",
          "Out of bounds access '%d'", j);
      else
        IoState_error_description_(IOSTATE, m, "Array.atPut",
          "Out of bounds access '%d'", j);
    }
    if (d == 0)
      i = j;
    else
      i += DATA(self)->idxCache[d - 1] * j;
  }

  val = NULL;
  if (at == 1)
  {
    val = DATA(self)->backingStore[i]; 
    return (val != NULL) ? val :  IONIL(self);
  }
  else
  {
    newVal = IoMessage_locals_objectArgAt_(m, locals, DATA(self)->d);
    DATA(self)->backingStore[i] = newVal;
    return newVal;
  }

  /* NOT REACHED. */
  return NULL;
}

IoArray* IoArray_proto(void *state)
{
  IoMethodTable methodTable[] = {
    {"sizeTo", IoArray_sizeTo_},
    {"at", IoArray_at},
    {"atPut", IoArray_atPut_},
    {NULL, NULL}
  };

  IoObject *self = IoObject_new(state);
  self->tag = IoArray_tag(state);
  self->data = calloc(1, sizeof(IoArrayData));

  IoState_registerProtoWithFunc_(state, self, IoArray_proto);
  IoObject_addMethodTable_(self, methodTable);

  return self;
}

IoArray* IoArray_rawClone(IoArray *self)
{
  IoObject *child = IoObject_rawClonePrimitive(self);
  IoArrayData *data;

  child->data = cpalloc(self->data, sizeof(IoArrayData));

  if (DATA(self)->d > 0)
  {
    data = (IoArrayData *) child->data;
    data->size = cpalloc(DATA(self)->size, sizeof(long) * DATA(self)->d);
    data->idxCache = cpalloc(DATA(self)->idxCache, sizeof(long) * DATA(self)->d);
    data->backingStore = cpalloc(DATA(self)->backingStore, sizeof(IoObject *) * DATA(self)->elems);
  }
  return child;
}

IoArray* IoArray_new(void *state)
{
  IoObject *proto = IoState_protoWithInitFunction_(state, IoArray_proto);
  return IoObject_rawClone(proto);
}

void IoArray_free(IoArray *self)
{
  free(DATA(self)->size);
  free(DATA(self)->idxCache);
  free(DATA(self)->backingStore);
  free(DATA(self));
}

void IoArray_mark(IoArray *self)
{
  long l;
  for (l = 0; l < DATA(self)->elems; l++)
  {
    if (DATA(self)->backingStore[l])
      IoObject_makeGrayIfWhite(DATA(self)->backingStore[l]);
  }
}

/*
 * Array sizeTo(d, n(0), n(1), ..., n(d-1))
 *  d: Dimentions.
 *  n(0) - n(d - 1): Size.
 *
 * Eg:
 *  foo := Array clone sizeTo(2, 2, 3);    # Create a 2x3 array.
 *  foo := Array clone sizeTo(3, 2, 3, 4); # Create a 2x3x4 array.
 */
IoArray* IoArray_sizeTo_(IoArray *self, IoObject *locals, IoMessage *m)
{
  int i;
  long j;
  long n = 0;
  long d = IoMessage_argCount(m);

  if (d < 0)
  {
    IoState_error_description_(IOSTATE, m, "Array.sizeTo",
      "Invalid d parameter '%d'", d);
  }

  DATA(self)->d = d;
  DATA(self)->size = calloc(d, sizeof(unsigned long));
  DATA(self)->idxCache = calloc(d, sizeof(unsigned long));
  for (i = 0; i < d; i++)
  {
    j = IoMessage_locals_longArgAt_(m, locals, i);
    DATA(self)->size[i] = j;
    if (i == 0) {
      n = j;
      DATA(self)->idxCache[i] = j;
    }
    else {
      DATA(self)->idxCache[i] = DATA(self)->idxCache[i - 1] * j;
      n = n * j;
    }
  }

  DATA(self)->elems = n;
  DATA(self)->backingStore = calloc(n, sizeof(IoObject *));
  if (DATA(self)->backingStore == NULL)
  {
    IoState_error_description_(IOSTATE, m, "Array.sizeTo",
      "Exhausted memory");
  }

  return self;
}

IoObject* IoArray_at(IoArray *self, IoObject *locals, IoMessage *m)
{
  return IoArray_handleAt(self, locals, m, 1);
}

IoObject* IoArray_atPut_(IoArray *self, IoObject *locals, IoMessage *m)
{
  return IoArray_handleAt(self, locals, m, 0);
}
