/*
 docCopyright("Steve Dekorte", 2002)
 docLicense("BSD revised")
 */

#include "IoMessage_parser.h"
#include "IoObject.h"
#include "IoSeq.h"
#include "IoMap.h"
#include "IoNumber.h"
#include "IoState.h"
#include "IoLexer.h"
#include "IoToken_parser.h"
#include <ctype.h>

#define DATA(self) ((IoMessageData *)IoObject_dataPointer(self))

void IoMessage_ifPossibleCacheToken_(IoMessage *self, IoToken *p) 
{
	IoSymbol *method = DATA(self)->name;
	IoObject *r = NULL;
	
	switch ((int)IoToken_type(p))
	{
		case TRIQUOTE_TOKEN:
			r =  IoState_symbolWithCString_length_(IOSTATE, IoSeq_asCString(method) + 3, IoSeq_rawSize(method) - 6); 
			break;
			
		case MONOQUOTE_TOKEN:
			r =  IoSeq_rawAsUnescapedSymbol(IoSeq_rawAsUnquotedSymbol(method)); 
			break;
			
		case NUMBER_TOKEN:
			r =  IONUMBER(IoSeq_asDouble(method));
			break;
			
		case HEXNUMBER_TOKEN:
			r =  IONUMBER(IoSeq_rawAsDoubleFromHex(method));
			break;      
			
		default:
			if (IoSeq_rawEqualsCString_(method, "Nil") || IoSeq_rawEqualsCString_(method, "nil"))
			{ 
				r = IONIL(self); 
			}
	}
	
	IoMessage_cachedResult_(self, r);
}

IoMessage *IoMessage_newFromText_label_(void *state, const char *text, const char *label)
{
	IoLexer *lexer = IoLexer_new();
	IoMessage *msg;
	
	IoLexer_string_(lexer, text);
	IoLexer_lex(lexer);
	
	
	msg = IoMessage_newParse(state, lexer);
	
	IoMessage_opShuffle_(msg);
	
	{
		IoSymbol *labelSymbol = IoState_symbolWithCString_((IoState *)state, label);
		IoMessage_label_(msg, labelSymbol); 
	}
	
	IoLexer_free(lexer);
	
	return msg;
}

// -------------------------------

int IoToken_LevelForOp_(char *s);

IoMessage *IoMessage_newParseNextMessageChain(void *state, IoLexer *lexer);

IoMessage *IoMessage_newParse(void *state, IoLexer *lexer)
{
	if (IoLexer_errorToken(lexer))
	{ 
		IoMessage *m; 
		IoSymbol *errorString;
		
                // Maybe the nil message could be used here. Or even a NULL.
                IoSymbol *error = IoState_symbolWithCString_(state, "Error");
		m = IoMessage_newWithName_returnsValue_(state, error, error);
		errorString = IoState_symbolWithCString_((IoState *)state, IoLexer_errorDescription(lexer));
		IoState_error_(state, m, "compile error: %s", CSTRING(errorString));
	}
	
	if (IoLexer_topType(lexer) == TERMINATOR_TOKEN)
	{
		IoLexer_pop(lexer);
	}
	
	if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
	{
		IoMessage *self = IoMessage_newParseNextMessageChain(state, lexer);
		
		if (IoLexer_topType(lexer) != NO_TOKEN)
		{
			// TODO: Exception as the end was expected
			IoState_error_(state, self, "compile error: %s", "unused tokens");
		}
		
		return self;
	}
	
	return IoMessage_newWithName_returnsValue_(state,
											   IoState_symbolWithCString_((IoState*)state, "nil"),
											   ((IoState*)state)->ioNil
											   );
}

IoMessage *IoMessage_newParseNextMessageChain(void *state, IoLexer *lexer)
{
	IoMessage *self = IoMessage_new(state);
	
	
	if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
	{
		IoMessage_parseName(self, lexer);
	}
	
	if (IoLexer_topType(lexer) == OPENPAREN_TOKEN)
	{
		IoMessage_parseArgs(self, lexer);
	}
	
	if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
	{
		IoMessage_parseNext(self, lexer);
	}
	
	while (IoLexer_topType(lexer) == TERMINATOR_TOKEN)
	{
		IoLexer_pop(lexer);
		
		if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
		{
			IoMessage *eol = IoMessage_newWithName_(state, ((IoState*)state)->semicolonSymbol);
			IoMessage_rawSetNextMessage(self, eol);
			IoMessage_parseNext(eol, lexer);
		}
	}
	
	return self;
}

void IoMessage_parseName(IoMessage *self, IoLexer *lexer)
{
	IoToken *token = IoLexer_pop(lexer);
	
	DATA(self)->name = IOREF(IOSYMBOL(IoToken_name(token)));
	
	IoMessage_ifPossibleCacheToken_(self, token);
	IoMessage_rawSetLineNumber_(self, IoToken_lineNumber(token));
	IoMessage_rawSetCharNumber_(self, IoToken_charNumber(token));
}

void IoMessage_parseArgs(IoMessage *self, IoLexer *lexer)
{
	IoLexer_pop(lexer);
	
	if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
	{
		IoMessage *arg = IoMessage_newParseNextMessageChain(IOSTATE, lexer);
		IoMessage_addArg_(self, arg);
		
		while (IoLexer_topType(lexer) == COMMA_TOKEN)
		{
			IoLexer_pop(lexer);
			
			if (IoTokenType_isValidMessageName(IoLexer_topType(lexer)))
			{
				IoMessage *arg = IoMessage_newParseNextMessageChain(IOSTATE, lexer);
				IoMessage_addArg_(self, arg);
			}
			// Doesn't actually work because the lexer detects this case and reports the error before we can handle it...
			//else if (IoLexer_topType(lexer) == CLOSEPAREN_TOKEN)
			//{
			//	// Allow the last arg to be empty as in, "foo(a,b,c,)".
			//}
				else
				{
					// TODO: Exception, missing message
				}
		}
	}
	
	if (IoLexer_topType(lexer) != CLOSEPAREN_TOKEN)
	{
		// TODO: Exception, missing close paren
	}
	IoLexer_pop(lexer);
}

void IoMessage_parseNext(IoMessage *self, IoLexer *lexer)
{
	IoMessage *next = IoMessage_newParseNextMessageChain(IOSTATE, lexer);
	IoMessage_rawSetNextMessage(self, next);
}


// --- Level ----------------------------------------------------------

// TODO move the IO_OP_MAX_LEVEL define to be adjacent to the op table.
// TODO Set IO_OP_MAX_LEVEL to max-1 or something massive otherwise people playing with custom operators don't have many levels available
#define IO_OP_MAX_LEVEL 17

enum LevelType {ATTACH, ARG, NEW, UNUSED};

typedef struct {
	IoMessage *message;
	enum LevelType type;
	int precedence;
} Level;

void Level_finish(Level *self);
void Level_attachAndReplace(Level *self, IoMessage *msg);
void Level_setAwaitingFirstArg(Level *self, IoMessage *msg, int precedence);

typedef struct {
	Level pool[IO_OP_MAX_LEVEL];
	int currentLevel;
	
	List *stack;
	IoMap *operatorTable;
} Levels;

void Levels_reset(Levels *self)
{
	int i;
	self->currentLevel = 1;
	
	for (i = 0; i < IO_OP_MAX_LEVEL; i ++)
	{
		Level *level = &self->pool[i];
		level->type = UNUSED;
	}
	
	{
		Level *level = &self->pool[0];
		level->message = NULL;
		level->type = NEW;
		level->precedence = IO_OP_MAX_LEVEL;
	}
	
	List_removeAll(self->stack);
	List_append_(self->stack, &self->pool[0]);
}

// --- Levels ----------------------------------------------------------

Levels *Levels_new(IoMessage *msg)
{
	Levels *self = calloc(1, sizeof(Levels));
	IoObject *opTable = IoObject_rawGetSlot_(msg, IoState_symbolWithCString_(msg->state, "OperatorTable"));
	
	self->operatorTable = NULL;
	
	if (opTable)
	{
		IoObject *rightOperators = IoObject_rawGetSlot_(opTable, IoState_symbolWithCString_(msg->state, "rightOperators"));
		
		if (rightOperators)
		{
			if (ISMAP(rightOperators))
			{
				self->operatorTable = rightOperators;
			}
			else
			{
				IoState_error_(msg->state, NULL, "compile error: Message OperatorTable rightOperators is not a Map. %p %s", rightOperators, CSTRING(IoMessage_name(msg)));
			}
		}
	}
	
	self->stack = List_new();
	Levels_reset(self);
	return self;
}

void Levels_free(Levels *self)
{
        List_free(self->stack);
	free(self);
}

inline Level *Levels_currentLevel(Levels *self)
{
	return List_top(self->stack);
}

void Levels_popDownTo(Levels *self, int targetLevel)
{
	Level *level;
	
	while (level = List_top(self->stack), level->precedence <= targetLevel && level->type != ARG)
	{
		Level_finish(List_pop(self->stack));
		self->currentLevel--;
	}
}

void Levels_attachToTopAndPush(Levels *self, IoMessage *msg, int precedence)
{
	Level *level = NULL;
	{
		Level *top = List_top(self->stack);
		Level_attachAndReplace(top, msg);
	}
	
	{
		// TODO: Check for overflow of the pool.
		if (self->currentLevel >= IO_OP_MAX_LEVEL)
		{
			IoState_error_(msg->state, NULL, "compile error: Overflowed operator stack. Only %d levels of operators currently supported.", IO_OP_MAX_LEVEL-1);
		}
		
		level = &self->pool[self->currentLevel++];
		Level_setAwaitingFirstArg(level, msg, precedence);
		List_append_(self->stack, level);
	}
}

void Level_finish(Level *self)
{
	if (self->message)
	{
		IoMessage_rawSetNextMessage(self->message, NULL);
	}
	
	self->type = UNUSED;
}

void Level_attach(Level *self, IoMessage *msg)
{
	switch (self->type)
	{
		case ATTACH:
			IoMessage_rawSetNextMessage(self->message, msg);
			break;
			
		case ARG:
			IoMessage_addArg_(self->message, msg);
			break;
			
		case NEW:
			self->message = msg;
			break;
			
		case UNUSED:
			break;
	}
}

void Level_attachAndReplace(Level *self, IoMessage *msg)
{
	Level_attach(self, msg);
	self->type = ATTACH;
	self->message = msg;
}

void Level_setAwaitingFirstArg(Level *self, IoMessage *msg, int precedence)
{
	self->type = ARG;
	self->message = msg;
	self->precedence = precedence;
}

void Level_setAlreadyHasArgs(Level *self, IoMessage *msg)
{
	self->type = ATTACH;
	self->message = msg;
}

int Levels_levelForOp(Levels *self, char *messageName, IoSymbol *messageSymbol, IoMessage *msg)
{
	if (self->operatorTable)
	{
		IoObject *value = IoMap_rawAt(self->operatorTable, messageSymbol);
		
		if (!value)
		{
			return -1;
		}
		
		if (ISNUMBER(value))
		{
			return IoNumber_asInt((IoNumber*)value);
		}
		else
		{
			IoState_error_(msg->state, msg, "compile error: Value for '%s' in Message OperatorTable is not a number. Values in the OperatorTable are numbers which indicate the precedence of the operator.", messageName);
			return -1; // To keep the compiler happy.
		}
	}
	else
	{
		return IoToken_LevelForOp_(messageName);
	}
}

void Levels_attach(Levels *self, IoMessage *msg, List *expressions)
{
	// TODO clean up this method.
	
	IoSymbol *messageSymbol = IoMessage_name(msg);
	char *messageName = CSTRING(messageSymbol);
	int precedence = Levels_levelForOp(self, messageName, messageSymbol, msg);
	
	// `o a := b c ; d`  becomes  `o setSlot("a", b c) ; d`
	// 
	// a      attaching
	// :=     msg
	// b c    msg->next
	
	if (strcmp(messageName, ":=") == 0 || strcmp(messageName, "=") == 0)
	{
		IoState *state = (IoState *)(msg->state);
		Level *currentLevel = Levels_currentLevel(self);
		IoMessage *attaching = currentLevel->message;
		IoSymbol *setSlotName;
		
		if (attaching == NULL) // `:= b ;`
		{
			// Could be handled as, message(:= 42) -> setSlot(nil, 42)
			
			IoState_error_(state, msg, "compile error: %s requires a symbol to its left.", messageName);
		}
		
		if (IoMessage_argCount(attaching) > 0) // `a(1,2,3) := b ;`
		{
			IoState_error_(state, msg, "compile error: The symbol to the left of %s cannot have arguments.", messageName);
		}
		
		
		{
			// `a := b ;`
			IoSymbol *slotName = DATA(attaching)->name;
			IoSymbol *quotedSlotName = IoSeq_newSymbolWithFormat_(state, "\"%s\"", CSTRING(slotName));
			IoMessage *slotNameMessage = IoMessage_newWithName_returnsValue_(state, quotedSlotName, slotName);
			
			// `a := b ;`  ->  `a("a") := b ;`
			IoMessage_addArg_(attaching, slotNameMessage);
			
			if (strcmp(messageName, ":=") == 0)
			{
				if (isupper(CSTRING(slotName)[0]))
				{
					setSlotName = state->setSlotWithTypeSymbol;
				}
				else
				{
					setSlotName = state->setSlotSymbol;
				}
			}
			else
			{
				setSlotName = state->updateSlotSymbol;
			}
		}
		
		// `a("a") := b ;`  ->  `setSlot("a") := b ;`
		DATA(attaching)->name = IoObject_addingRef_(attaching, setSlotName);
		
		currentLevel->type = ATTACH;
		
		if (IoMessage_argCount(msg) > 0) // `setSlot("a") :=(b c) d e ;`
		{
			// `b c`
			IoMessage *arg = IoMessage_rawArgAt_(msg, 0);
			
			// `()`
			IoMessage *foo = IoMessage_newWithName_(state, IoState_symbolWithCString_(state, ""));
			
			// `()`  ->  `(b c)`
			IoMessage_addArg_(foo, arg);
			
			// `(b c)`  ->  `(b c) d e ;`
			IoMessage_rawSetNextMessage(foo, DATA(msg)->next);
			
			// `setSlot("a") :=(b c) d e ;`  ->  `setSlot("a", (b c) d e ;) :=(b c) d e ;`
			IoMessage_addArg_(attaching, foo);
		}
		else // `setSlot("a") := b ;`
		{
			// `setSlot("a") :=` or `setSlot("a") := ;`
			IoMessage *mn = DATA(msg)->next;
			IoSymbol *name = mn ? DATA(mn)->name : NULL;
			IoSymbol *semi = ((IoState *)(msg->tag->state))->semicolonSymbol;
			
			//if (mn == NULL || IoMessage_isEOL(mn))
			if (mn == NULL || name == semi)
			{
				IoState_error_(state, msg, "compile error: %s must be followed by a value.", messageName);
			}
			
			// `setSlot("a") := b c ;`  ->  `setSlot("a", b c ;) := b c ;`
			IoMessage_addArg_(attaching, DATA(msg)->next);
		}
		
		// process the value (`b c d`) later  (`setSlot("a", b c d) := b c d ;`)
		if (DATA(msg)->next != NULL && !IoMessage_isEOL(DATA(msg)->next))
		{
			List_push_(expressions, DATA(msg)->next);
		}
		
		{
			IoMessage *last = msg;
			while (DATA(last)->next != NULL && !IoMessage_isEOL(DATA(last)->next))
			{
				last = DATA(last)->next;
			}
			
			IoMessage_rawSetNextMessage(attaching, DATA(last)->next);
			
			// Continue processing in IoMessage_opShuffle loop
			IoMessage_rawSetNextMessage(msg, DATA(last)->next);
			
			IoMessage_rawSetNextMessage(last, NULL);
		}
		
		// make sure b in `1 := b` gets executed
		IoMessage_cachedResult_(attaching, NULL);
	}
	else if (IoMessage_isEOL(msg))
	{
		Levels_popDownTo(self, IO_OP_MAX_LEVEL-1);
		Level_attachAndReplace(Levels_currentLevel(self), msg);
	}
	else if (precedence != -1)
	{
		if (IoMessage_argCount(msg) > 0)
		{
			Level_setAlreadyHasArgs(Levels_currentLevel(self), msg);
		}
		else
		{
			Levels_popDownTo(self, precedence);
			Levels_attachToTopAndPush(self, msg, precedence);
		}
	}
	else
	{
		Level_attachAndReplace(Levels_currentLevel(self), msg);
	}
}

void Levels_nextMessage(Levels *self)
{
	Level *level;
	
	while ((level = List_pop(self->stack)))
	{
		Level_finish(level);
	}
	
	Levels_reset(self);
}

void IoMessage_opShuffle_(IoMessage *self)
{
	if (IoObject_rawGetSlot_(self, IOSTATE->opShuffleSymbol) && IoMessage_name(self) != IOSTATE->noShufflingSymbol)
	{
		IoMessage_locals_performOn_(IOSTATE->opShuffleMessage, IOSTATE->lobby, self);
	}
}

IoMessage *IoMessage_opShuffle(IoMessage *self, IoObject *locals, IoMessage *m)
{
	Levels *levels = Levels_new(self);
	List *expressions = List_new();
	
	List_push_(expressions, self);
	
	while (List_size(expressions) >= 1)
	{
		IoMessage *n = List_pop(expressions);
		
		do
		{
			Levels_attach(levels, n, expressions);
			List_appendSeq_(expressions, DATA(n)->args);
		} while (n = DATA(n)->next);
		
		Levels_nextMessage(levels);
	}

        List_free(expressions);
	Levels_free(levels);

	return self;
}

